import numpy as np
from colorama import Fore, Style, init
import os
from collections import deque
from enum import Enum
init()

class SolveError(ValueError):
    pass

class SolveMethods(Enum):
    CELL_ELIMINATION = 1
    SEGMENT_ELIMINATION = 2
    GUESS = 3

class SudokuBoard:
    # BOARD_VALUES = np.array(range(1,10),dtype=np.uint8)
    ROWS = np.array(range(0,9),dtype=np.uint8)
    COLUMNS = np.array(range(0,9),dtype=np.uint8)
    SIZE = np.array([len(ROWS),len(COLUMNS)])
    X, Y = np.meshgrid(ROWS,COLUMNS)
    CELLS = np.transpose(np.concat((np.expand_dims(X,2),np.expand_dims(Y,2)),axis=2),(2,0,1))
    VALUES = {}
    NUMBERS = {}
    VALUE_BITS = {}
    VALUE_MAP = {}
    VALUE_MASK = 2**(SIZE[0]+1)-1
    
    def __init__(self,initialValues=None,finalValues=None):
        self.board = np.array([[None for i in range(SudokuBoard.SIZE[0])] for i in range(SudokuBoard.SIZE[1])])
        self.segments = []

        SudokuBoard.__initValueBits()
        self._initCells(initialValues = initialValues, finalValues = finalValues)
        self._initSegments()

    @staticmethod
    def __initValueBits():
        for i in range(SudokuBoard.SIZE[0]+1):
            SudokuBoard.VALUES[i] = np.uint16(2**i)
            SudokuBoard.NUMBERS[2**i] = np.uint16(i)

        for i in range(2**(SudokuBoard.SIZE[0]+1)):
            bits = bin(i)[::-1]
            ind = np.array([j for j, ltr in enumerate(bits) if ltr == '1'],dtype=np.uint16)
            SudokuBoard.VALUE_MAP[i] = np.pow(2, ind)
            SudokuBoard.VALUE_BITS[i] = len(SudokuBoard.VALUE_MAP[i])
        pass

    @staticmethod
    def decomposeValue(value):
        return SudokuBoard.VALUE_BITS[value]
    
    @staticmethod
    def composeValues(values):
        return np.bitwise_or.reduce(values)
    
    @staticmethod
    def invertValue(value):
        return np.bitwise_and(np.bitwise_not(value), SudokuBoard.VALUE_MASK)
    
    @staticmethod
    def combineValues(value1,value2):
        return np.bitwise_or(value1,value2)
    
    @staticmethod
    def removeValue(value1,value2):
        return np.bitwise_and(value1,np.bitwise_not(value2))
    
    def check(self):
        for segment in self.segments:
            invalid = segment.check()
            if invalid is not None:
                cell1,cell2 = invalid
                print(f'Invalid value set at {cell1.coord} and {cell2.coord} with values {cell1.values} and {cell2.values}')
                return cell1,cell2
        return None
    
    @property
    def valid(self):
        return self.check() is None
    
    @property
    def solvedValues(self):
        solvedValues = [[0 for i in range(SudokuBoard.SIZE[0])] for i in range(SudokuBoard.SIZE[1])]
        for row in range(SudokuBoard.SIZE[0]):
            for col in range(SudokuBoard.SIZE[1]):
                solvedValues[row][col] = int(self.board[row,col].solvedValue)
        return solvedValues
    
    @property
    def solvedNumbers(self):
        solvedNumbers = [[0 for i in range(SudokuBoard.SIZE[0])] for i in range(SudokuBoard.SIZE[1])]
        for row in range(SudokuBoard.SIZE[0]):
            for col in range(SudokuBoard.SIZE[1]):
                solvedNumbers[row][col] = int(self.board[row,col].solvedValue)
        return solvedNumbers
    
    @property
    def values(self):
        values = [[0 for i in range(SudokuBoard.SIZE[0])] for i in range(SudokuBoard.SIZE[1])]
        for row in range(SudokuBoard.SIZE[0]):
            for col in range(SudokuBoard.SIZE[1]):
                values[row][col] = int(self.board[row,col].value)
        return values
    
    @property
    def solved(self):
        for row in range(SudokuBoard.SIZE[0]):
            for col in range(SudokuBoard.SIZE[1]):
                if not self.board[row,col].solved:
                    return False
        return True

    
    def getState(self):
        cellValues = [[0 for i in range(SudokuBoard.SIZE[0])] for i in range(SudokuBoard.SIZE[1])]
        for row in range(SudokuBoard.SIZE[0]):
            for col in range(SudokuBoard.SIZE[1]):
                cellValues[row][col] = int(self.board[row,col].value)
        segmentValues = [0 for _ in self.segments]
        for i, segment in enumerate(self.segments):
            segmentValues[i] = segment.value
        return cellValues, segmentValues

    def setState(self, cellValues, segmentValues):
        for row in range(SudokuBoard.SIZE[0]):
            for col in range(SudokuBoard.SIZE[1]):
                self.board[row,col].value = cellValues[row][col]
        for i, segment in enumerate(self.segments):
            segment.value = segmentValues[i]

    def _initCells(self,initialValues,finalValues):
        for row in SudokuBoard.ROWS:
            for column in SudokuBoard.COLUMNS:
                cell = SudokuBoardCell(row,column,self)
                if initialValues is not None:
                    cell.value = SudokuBoard.VALUES[initialValues[row][column]]
                    cell.initial = cell.value != SudokuBoard.VALUES[0]
                    if finalValues is not None:
                        cell.truth = SudokuBoard.VALUES[finalValues[row][column]]
                self.board[row,column] = (cell)

        pass

    def _initSegments(self):

        # Init rows
        for row in SudokuBoard.ROWS:
            cells = self.board[row,:]
            segment = SudokuBoardSegment(cells,self)
            self.segments.append(segment)

        # Init columns
        for column in SudokuBoard.COLUMNS:
            cells = self.board[:,column]
            segment = SudokuBoardSegment(cells,self)
            self.segments.append(segment)

        # Init diagonals
        # segment = SudokuBoardSegment(self.board.diagonal(),self)
        # self.segments.append(segment)
        # segment = SudokuBoardSegment(np.fliplr(self.board).diagonal(),self)
        # self.segments.append(segment)

        # Init squares
        for row in range(0,3):
            for column in range(0,3):
                cells = self.board[row*3:row*3+3,column*3:column*3+3]
                segment = SudokuBoardSegment(cells.flatten(),self)
                self.segments.append(segment)

    def solve(self):
        changed = True
        iteration = 1
        guesses = deque()

        while not self.solved and changed:

            try:
                # Solve segments one iteration
                for segment in self.segments:
                    yield from segment.solve()

                # Solve cells until stuck
                while changed:
                    iteration += 1
                    changed = False

                    # Solve cells one iteration
                    for row in range(SudokuBoard.SIZE[0]):
                        for column in range(SudokuBoard.SIZE[1]):
                            for s,c,m in self.board[row,column].solve():
                                changed = changed or c
                                if c:
                                    yield s,c,m

                    # Solve segments one iteration
                    for segment in self.segments:
                        for s,c,m in segment.solve():
                            changed = changed or c
            
                # Got stuck. Choose a random value
                for row in range(SudokuBoard.SIZE[0]):
                    for column in range(SudokuBoard.SIZE[1]):
                        if self.board[row,column].bits == 2:
                            guessCell = self.board[row,column]
                            guessValue = guessCell.values[0]
                            cellValues, segmentValues = self.getState()
                            guess = (guessCell, guessValue, cellValues, segmentValues)
                            guesses.append(guess)
                            guessCell.value = guessValue
                            changed = True
                            print(f'Guessing {guessValue} for {guessCell.coord}')
                            yield self.board[row,column].solved, changed, SolveMethods.GUESS
                            break
                    if changed:
                        break

            except SolveError:
                # Had a bad guess, try again
                # print('Solve Error')
                changed = False
                while len(guesses) > 0 and not changed:
                    # print('Bad Guess, retrying')
                    # Restore previous state
                    guessCell, guessValue, cellValues, segmentValues = guesses.pop()
                    self.setState(cellValues, segmentValues)

                    # Formulate new guess
                    guessIndex = guessCell.values.index(guessValue) + 1
                    if guessIndex < len(guessCell.values):
                        guessValue = guessCell.values[guessIndex]
                        guess = (guessCell, guessValue, cellValues, segmentValues)
                        guesses.append(guess)
                        guessCell.value = guessValue
                        changed = True
                        yield guessCell.solved, changed, SolveMethods.GUESS

                if not changed:
                    raise ValueError('No solution found')

        pass

    def format(self):
        table = [['']*SudokuBoard.SIZE[1] for i in range(SudokuBoard.SIZE[0])]
        columnWidths = [0]*SudokuBoard.SIZE[1]
        for row in range(SudokuBoard.SIZE[0]):
            for column in range(SudokuBoard.SIZE[1]):
                table[row][column] = str(self.board[row,column]).replace(' ','')
                columnWidths[column] = max(columnWidths[column],len(table[row][column]))
                
        txt = '\n| '
        for row in range(SudokuBoard.SIZE[0]):
            for column in range(SudokuBoard.SIZE[1]):
                txt += table[row][column].ljust(columnWidths[column]+1) + ' | '
            txt += '\n| '

        return txt

    def __str__(self):
        table = [['']*SudokuBoard.SIZE[1] for i in range(SudokuBoard.SIZE[0])]
        colors = [['']*SudokuBoard.SIZE[1] for i in range(SudokuBoard.SIZE[0])]
        columnWidths = [0]*SudokuBoard.SIZE[1]
        for row in range(SudokuBoard.SIZE[0]):
            for column in range(SudokuBoard.SIZE[1]):
                table[row][column] = str(self.board[row,column]).replace(' ','')
                if self.board[row][column].changed:
                    colors[row][column] = Fore.RED
                elif self.board[row][column].tested:
                    colors[row][column] = Fore.BLUE
                elif self.board[row][column].initial:
                    colors[row][column] = Fore.YELLOW
                elif self.board[row][column].solved:
                    colors[row][column] = Fore.GREEN
                columnWidths[column] = max(columnWidths[column],len(table[row][column]))
                
        txt = '\n| '
        for row in range(SudokuBoard.SIZE[0]):
            for column in range(SudokuBoard.SIZE[1]):
                txt += colors[row][column] + table[row][column].ljust(columnWidths[column]+1) + Style.RESET_ALL + ' | '
            txt += '\n| '

        return txt

class SudokuBoardSegment:
    def __init__(self, cells, board):
        self._board = board
        self._cells = []
        self._value = 0
        self._changed = False
        self._tested = False
        for cell in cells:
            self.addCell(cell)

    def getValues(self):
        return [cell.value for cell in self._cells]
    
    @property
    def value(self):
        return self._value
    
    @value.setter
    def value(self, value):
        self._changed = self.value != value
        self._value = value
    
    @property
    def solvedValue(self):
        self.tested = True
        return self.value

    @property
    def board(self):
        return self._board
    
    @property
    def cells(self):
        return self._cells
    
    @property
    def tested(self):
        tested = self._tested
        self.tested = False
        return tested

    @tested.setter
    def tested(self,value):
        self._tested = value
        for cell in self._cells:
            cell.tested = value
    
    def addCell(self, cell):
        if cell not in self._cells:
            self._cells.append(cell)
            cell.addSegment(self)

    @property
    def solved(self):
        return self.bits == 9
    
    @property
    def changed(self):
        changed = self._changed
        self._changed = False
        return changed

    @property
    def bits(self):
        return SudokuBoard.VALUE_BITS[self.value]
    
    def check(self):
        for cell1 in self._cells:
            if cell1.solvedValue != 1:
                for cell2 in self._cells:
                    if cell1 != cell2 and cell2.solvedValue != 1 and cell1.solvedValue == cell2.solvedValue:
                        return (cell1,cell2)
        return None

    @property
    def valid(self):
        return self.check is None

    def solve(self):
        if self.solved:
            return self.solved, self.changed
        
        value = np.uint16(0)
        values = {}
        changed = False
        for cell in self._cells:
            for val in cell.values:
                if val not in values:
                    values[val] = [cell]
                else:
                    values[val].append(cell)
            value = SudokuBoard.combineValues(value,cell.solvedValue)
        if SudokuBoard.VALUES[0] not in values:
            for val, cells in values.items():
                if len(cells) == 1 and cells[0].value != val:
                    cells[0].value = val
                    changed = True
                    # Only do once to ensure each selection has time to propogate through the solution
                    yield cells[0].solved, changed, SolveMethods.SEGMENT_ELIMINATION
                    break

        self.value = value
        if self.check() is not None:
            raise SolveError(f'Segment has duplicate values\n{self._board.format()}')
        yield self.solved, self._changed, None


class SudokuBoardCell:
    def __init__(self, row, column, board):
        self._board = board
        self._row = row
        self._column = column
        self._value = 0
        self._segments = []
        self._changed = False
        self._initial = False
        self._tested = False
        self.truth = None

    @property
    def board(self):
        return self._board

    @property
    def row(self):
        return self._row

    @property
    def column(self):
        return self._column
    
    @property
    def coord(self):
        return (self._row, self._column)
    
    @property
    def value(self):
        return self._value
    
    @value.setter
    def value(self, value):
        # print(f'Setting Cell ({int(self.coord[0])},{int(self.coord[1])}) to {SudokuBoard.VALUE_MAP[value]}')
        if value == 0:
            raise ValueError('Cell value cannot be empty')
        
        if self.solvedValue != 1 and self.truth is not None and self.truth not in self.values:
            raise ValueError(f'True value {self.truth} not in {list(SudokuBoard.VALUE_MAP[int(self.value)])}')
        
        self._changed = self.value != value
        self._value = value

        # for segment in self._segments:
        #     if segment.check() is not None:
        #         raise SolveError(f'Duplicate value in segment set for {self.coord}\n{self._board.format()}')
    
    @property
    def solvedValue(self):
        if self.solved:
            return self._value
        else:
            return 1
    @property
    def solved(self):
        return self.bits == 1 and self._value != 1
    
    @property
    def solvedNumber(self):
        return SudokuBoard.NUMBERS[self.solvedValue]

    @property
    def changed(self):
        changed = self._changed
        self._changed = False
        return changed

    @property
    def initial(self):
        return self._initial
    
    @initial.setter
    def initial(self, initial):
        self._initial = initial
    
    @property
    def tested(self):
        tested = self._tested and self.solved
        self._tested = False
        return tested

    @tested.setter
    def tested(self,value):
        self._tested = value
    
    def addSegment(self, segment):
        if segment not in self._segments:
            self._segments.append(segment)
            segment.addCell(self)

    def solve(self):
        if self.solved:
            return self.solved, self._changed, None
        
        blockedValue = 0
        for segment in self._segments:
            segmentValue = segment.solvedValue
            
            # if SudokuBoard.VALUE_BITS[segmentValue] == 8:
            #     blockedValue = SudokuBoard.invertValue(segmentValue)
            #     break
            blockedValue = SudokuBoard.combineValues(blockedValue,segmentValue)
        
        if blockedValue == list(SudokuBoard.VALUE_MAP.keys())[-1]:
            # print(self._board)
            raise SolveError(f'No solution found for {self.coord}\n{self._board.format()}')
        
        self.value = SudokuBoard.invertValue(blockedValue)
        # valid = self._board.check()
        # if not valid:
        #     print(self._board)
        #     pass
        changed = self._changed
        yield self.solved, self._changed, SolveMethods.CELL_ELIMINATION

        if changed:
            for segment in self._segments:
                yield from segment.solve()


    @property
    def values(self):
        return [int(i) for i in list(SudokuBoard.VALUE_MAP[int(self.value)])]
    
    @property
    def numbers(self):
        return [SudokuBoard.NUMBERS[i] for i in self.values]
    
    @property
    def bits(self):
        return SudokuBoard.VALUE_BITS[int(self._value)]

    def __str__(self):
        lst = [int(i) for i in self.numbers]
        ret = str(lst)
        if self.truth is not None:
            ret += f'({int(SudokuBoard.NUMBERS[int(self.truth)])})'
        return ret

if __name__ == '__main__':

    initValues1 = [
        [0,7,0,0,0,6,5,3,4],
        [0,5,1,0,0,4,0,6,0],
        [0,0,0,2,5,0,8,0,9],
        [0,6,0,0,7,9,0,4,0],
        [0,0,0,4,2,0,0,7,0],
        [7,4,0,0,8,0,0,5,0],
        [3,0,0,5,9,2,4,8,6],
        [0,9,8,0,0,3,0,0,0],
        [0,2,4,0,6,8,0,0,1]
    ]
    initValues1 = None
    
    initValues2 = [
        [0,8,1,0,3,0,9,0,0],
        [0,0,0,0,2,0,3,8,0],
        [3,2,6,9,7,0,1,4,5],
        [8,1,0,0,0,0,6,0,0],
        [6,0,0,8,0,7,0,9,0],
        [0,0,0,5,0,0,8,7,0],
        [0,6,0,7,5,4,0,0,0],
        [5,0,4,0,0,0,0,0,9],
        [2,7,8,1,0,0,0,6,4],
    ]

    finalValues2 = [
        [7,8,1,4,3,5,9,2,6],
        [9,4,5,6,2,1,3,8,7],
        [3,2,6,9,7,8,1,4,5],
        [8,1,7,3,4,9,6,5,2],
        [6,5,2,8,1,7,4,9,3],
        [4,9,3,5,6,2,8,7,1],
        [1,6,9,7,5,4,2,3,8],
        [5,3,4,2,8,6,7,1,9],
        [2,7,8,1,9,3,5,6,4],
    ]

    initValues3 = [
        [0,0,0,0,4,0,0,2,6],
        [3,0,9,7,2,0,0,0,0],
        [0,5,0,0,0,0,4,0,0],
        [0,0,7,9,1,3,0,0,0],
        [6,0,3,0,0,0,0,0,0],
        [0,0,0,0,0,0,0,0,2],
        [1,0,0,0,0,0,0,8,0],
        [0,0,0,0,0,8,0,0,9],
        [0,9,0,0,0,0,7,0,4],
    ]
    finalValues3 = None

    initValues4 = [
        [0,0,8,0,4,0,0,0,0],
        [0,0,2,0,0,1,4,0,3],
        [0,0,0,0,0,0,8,0,0],
        [0,0,0,9,0,0,5,6,0],
        [5,0,0,0,8,7,0,0,0],
        [0,0,0,2,0,0,0,9,0],
        [0,3,1,0,0,0,0,0,0],
        [0,9,0,0,0,2,7,0,0],
        [0,0,7,5,0,0,0,0,9],
    ]
    finalValues4 = None


    initValues = initValues4
    finalValues = finalValues4
    if initValues is not None and finalValues is not None:
        for row in range(9):
            for col in range(9):
                if initValues[row][col] != 0 and initValues[row][col] != finalValues[row][col]:
                    raise ValueError(f'Incorrect value at ({row},{col}): {initValues[row][col]} != {finalValues[row][col]} at ({row},{col})')
                
    board = SudokuBoard(initialValues = initValues, finalValues = finalValues)
    # final = SudokuBoard(initialValues = finalValues2)

    str(board)
    print(board)

    # solved = False
    # changed = True
    # while not solved and changed:
    #     solved = True
    #     changed = False
    for s,c,m in board.solve():
        valid = board.valid
        print(m)
        print(board)
        # if not valid:
        #     break

        if c:
            # print('============================')
            # print(board)
            # print(final)
            pass
            # solved &= s
            # changed |= c

    print(f'========== Final ==========')
    str(board)
    print(board)
    print(board.valid)